"""CPython API Methods and intrinsic constants."""
from __future__ import annotations

import ctypes
from collections.abc import Sequence
from ctypes import Array, c_size_t, c_void_p, pythonapi, sizeof
from typing import Any, Callable, TypeVar

import _ctypes
from typing_extensions import Annotated

from einspect.compat import Version, python_req
from einspect.protocols.delayed_bind import bind_api

__all__ = (
    "Py",
    "Py_ssize_t",
    "Py_hash_t",
    "uintptr_t",
    "PTR_SIZE",
    "PyObj_FromPtr",
    "ALIGNMENT",
    "ALIGNMENT_SHIFT",
    "align_size",
    "address",
    "seq_to_array",
)

from einspect.types import NULL, Pointer

_T = TypeVar("_T")
_CT = TypeVar("_CT")

Py_ssize_t = ctypes.c_ssize_t
"""Constant for type Py_ssize_t."""

Py_hash_t = ctypes.c_ssize_t
"""Constant for type Py_hash_t."""

uintptr_t = ctypes.c_uint64
"""Constant for type uintptr_t."""

PTR_SIZE = sizeof(c_void_p)
"""Size of a pointer in bytes."""

# Alignments (must be powers of 2)
# https://github.com/python/cpython/blob/3.11/Objects/obmalloc.c#L878-L884
if sizeof(c_void_p) > 4:
    ALIGNMENT = 16
    ALIGNMENT_SHIFT = 4
else:  # pragma: no cover
    ALIGNMENT = 8
    ALIGNMENT_SHIFT = 3


class Py:
    """Typed methods from pythonapi."""

    @bind_api(pythonapi["Py_IncRef"])
    @staticmethod
    def IncRef(obj: object) -> None:
        """
        Increment the reference count of an object.

        https://docs.python.org/3/c-api/refcounting.html#c.Py_IncRef
        """

    @bind_api(pythonapi["Py_DecRef"])
    @staticmethod
    def DecRef(obj: object) -> None:
        """
        Decrements the reference count of an object.

        https://docs.python.org/3/c-api/refcounting.html#c.Py_DecRef
        """

    @bind_api(python_req(Version.PY_3_10) or pythonapi["Py_NewRef"])
    @staticmethod
    def NewRef(obj: object) -> object:
        """
        Return a new reference to an object and increment its refcount.

        https://docs.python.org/3/c-api/refcounting.html#c.Py_NewRef
        """

    class Tuple:
        @bind_api(pythonapi["PyTuple_Size"])
        @staticmethod
        def Size(obj: tuple) -> int:
            """
            Return the size of a tuple object.
            https://docs.python.org/3/c-api/tuple.html#c.PyTuple_Size
            """

        @bind_api(pythonapi["PyTuple_GetItem"])
        @staticmethod
        def GetItem(obj: tuple, index: int) -> object:
            """
            Return the item at position index in the tuple o.
            https://docs.python.org/3/c-api/tuple.html#c.PyTuple_GetItem
            """

        @bind_api(pythonapi["PyTuple_SetItem"])
        @staticmethod
        def SetItem(obj: tuple, index: int, value: object) -> None:
            """
            Set the item at position index in the tuple o to v.
            https://docs.python.org/3/c-api/tuple.html#c.PyTuple_SetItem

            Notes:
                - This function steals a reference to v.
                - Requires tuple o to have a reference count == 1.
            """

        @bind_api(pythonapi["_PyTuple_Resize"])
        @staticmethod
        def Resize(obj: tuple, size: int) -> None:
            """
            Resize the tuple to the specified size.
            https://docs.python.org/3/c-api/tuple.html#c._PyTuple_Resize

            Notes:
                - Not part of the documented limited C API.
                - Requires tuple o to have ref-count = 1 or size = 0.
            """

    class Type:
        @bind_api(pythonapi["PyType_Modified"])
        @staticmethod
        def Modified(obj: object) -> None:
            """
            Invalidate the internal lookup cache for the type and all of its subtypes.

            - This function must be called after any manual modification of the attributes
              or base classes of the type.
            """

    class Mem:
        @bind_api(pythonapi["PyMem_Malloc"])
        @staticmethod
        def Malloc(n: Annotated[int, c_size_t]) -> c_void_p:
            """
            Allocates n bytes and returns a pointer of type void* to the allocated memory,
            or NULL if the request fails.

            Requesting zero bytes returns a distinct non-NULL pointer if possible,
            as if PyMem_Malloc(1) had been called instead. The memory will not
            have been initialized in any way.

            https://docs.python.org/3/c-api/memory.htm
            """

        @bind_api(pythonapi["PyMem_Calloc"])
        @staticmethod
        def Calloc(
            nelem: Annotated[int, c_size_t], elsize: Annotated[int, c_size_t]
        ) -> c_void_p:
            """
            Allocates nelem elements each whose size in bytes is elsize and
            returns  The memory is initialized to zeros.

            Requesting zero elements or elements of size zero bytes
            returns a distinct non-NULL pointer if possible,
            as if PyMem_Calloc(1, 1) had been called instead.

            Args:
                nelem: Number of elements.
                elsize: Size of each element in bytes.

            Returns:
                A pointer to the allocated memory, or NULL if the request fails.
            """

        @bind_api(pythonapi["PyMem_Realloc"])
        @staticmethod
        def Realloc(p: c_void_p, n: Annotated[int, c_size_t]) -> c_void_p:
            """
            Resizes the memory block pointed to by p to n bytes.
            The contents will be unchanged to the minimum of the old and the new sizes.

            Args:
                p: Pointer to the memory block to be resized.
                n: New size in bytes.

            Returns:
                Pointer p, or NULL if request fails (p remains a valid pointer).
            """


PyObj_FromPtr: Callable[[int], object] = _ctypes.PyObj_FromPtr
"""(Py_ssize_t ptr) -> Py_ssize_t"""


def address(obj: Any) -> int:
    """Return the address of a python object. Same as id()."""
    source = ctypes.py_object(obj)
    addr = ctypes.c_void_p.from_buffer(source).value
    if addr is None:
        raise ValueError("address: NULL object")  # pragma: no cover
    return addr


def align_size(size: int, alignment: int = ALIGNMENT) -> int:
    """Align size to alignment."""
    return (size + alignment - 1) & ~(alignment - 1)


def seq_to_array(
    seq: Sequence[_T] | Array[_T], dtype: type[_CT], py_obj_try_cast: bool = True
) -> Array[_CT]:
    """Cast a Sequence to a ctypes.Array of a given type."""
    if isinstance(seq, Array):
        return seq

    arr_type = dtype * len(seq)

    if py_obj_try_cast and issubclass(dtype, Pointer):
        dtype_r = dtype._type_
        if hasattr(dtype_r, "try_from"):
            # If we find a NULL singleton, don't set it.
            arr = arr_type()
            for i, v in enumerate(seq):
                if v is not NULL:
                    arr[i] = dtype_r.try_from(v).with_ref().as_ref()
            return arr

    return arr_type(*seq)



from __future__ import annotations

from ctypes import POINTER, c_long, pythonapi
from typing import List, TypeVar

from typing_extensions import Annotated

from einspect.protocols.delayed_bind import bind_api
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyObject, PyVarObject
from einspect.structs.traits import IsGC
from einspect.types import ptr

_VT = TypeVar("_VT")


# noinspection PyPep8Naming
@struct
class PyListObject(PyVarObject[list, None, _VT], IsGC):
    """
    Defines a PyListObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/listobject.h
    """

    ob_item: ptr[ptr[PyObject[_VT, None, None]]]
    allocated: Annotated[int, c_long]

    @classmethod
    def from_object(cls, obj: list[_VT]) -> PyListObject[_VT]:
        return super().from_object(obj)  # type: ignore

    def _format_fields_(self) -> Fields:
        return {
            **super()._format_fields_(),
            "ob_item": ("**PyObject", ptr[ptr[PyObject] * self.ob_size]),
            "allocated": "c_long",
        }

    @bind_api(pythonapi["PyList_GetItem"])
    def GetItem(self, index: int) -> POINTER(PyObject):
        """
        Return the object at position index in the list pointed to by list.

        - If index is out of bounds (<0 or >=len(list)), return NULL and set an IndexError exception.
        - The position must be non-negative; indexing from the end of the list is not supported.

        Returns:
            pointer[PyObject] (borrowed reference) or pointer[NULL] on failure.
        """

    @bind_api(pythonapi["PyList_GetSlice"])
    def GetSlice(self, low: int, high: int) -> POINTER(PyListObject):
        """
        Return a list of the objects in list containing the objects between low and high.
        Analogous to list[low:high].

        - Return NULL and set an exception if unsuccessful.
        - Indexing from the end of the list is not supported.

        Returns:
            pointer[PyListObject] (new reference) or pointer[NULL] on failure.
        """

    @bind_api(pythonapi["PyList_SetItem"])
    def SetItem(self, index: int, value: object) -> None:
        """Set a value to a given index."""

    @bind_api(pythonapi["PyList_SetSlice"])
    def SetSlice(self, low: int, high: int, item_list: List[_VT]) -> None:
        """Set a value to a given index."""



from __future__ import annotations

from ctypes import c_uint32
from types import FunctionType

from typing_extensions import Annotated

from einspect.structs.deco import struct
from einspect.structs.include.object_h import vectorcallfunc
from einspect.structs.py_object import PyObject
from einspect.types import ptr


@struct
class PyFunctionObject(PyObject[FunctionType, None, None]):
    globals: ptr[PyObject]
    builtins: ptr[PyObject]
    name: ptr[PyObject]
    qualname: ptr[PyObject]
    code: ptr[PyObject]  # A code object, the __code__ attribute
    defaults: ptr[PyObject]  # NULL or a tuple
    kwdefaults: ptr[PyObject]  # NULL or a dict
    closure: ptr[PyObject]  # NULL or a tuple of cell objects

    func_doc: ptr[PyObject]  # The __doc__ attribute, can be anything
    func_dict: ptr[PyObject]  # The __dict__ attribute, a dict or NULL
    func_weakreflist: ptr[PyObject]  # List of weak references
    func_module: ptr[PyObject]  # The __module__ attribute, can be anything
    func_annotations: ptr[PyObject]  # Annotations, a dict or NULL
    vectorcall: vectorcallfunc

    func_version: Annotated[int, c_uint32]



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, Callable, Type, TypeVar

from typing_extensions import Self

from einspect.compat import Version
from einspect.errors import UnsafeError
from einspect.structs.include.object_h import TpFlags
from einspect.structs.py_type import PyTypeObject
from einspect.structs.slots_map import Slot, get_slot
from einspect.type_orig import add_cache, in_cache
from einspect.views.view_base import REF_DEFAULT, VarView

__all__ = ("TypeView", "impl")

MISSING = object()

_T = TypeVar("_T")
_Fn = TypeVar("_Fn", bound=Callable)


def impl(cls: Type[_T]) -> Callable[[_Fn], _Fn]:
    """Decorator for implementing methods on built-in types."""
    if not isinstance(cls, type):
        raise TypeError("cls must be a type")

    t_view = TypeView(cls)

    def wrapper(func: _Fn) -> _Fn:
        if isinstance(func, property):
            name = func.fget.__name__
        else:
            name = func.__name__

        t_view[name] = func
        return func

    return wrapper


class TypeView(VarView[_T, None, None]):
    _pyobject: PyTypeObject[_T]

    def __init__(self, obj: _T, ref: bool = REF_DEFAULT) -> None:
        """Create a new TypeView."""
        super().__init__(obj, ref)

    @property
    def immutable(self) -> bool:
        """Return True if the type is immutable."""
        if Version.PY_3_10.above():
            return bool(self._pyobject.tp_flags & TpFlags.IMMUTABLETYPE)
        return not bool(self._pyobject.tp_flags & TpFlags.HEAPTYPE)  # pragma: no cover

    @immutable.setter
    def immutable(self, value: bool):
        """Set whether the type is immutable."""
        if Version.PY_3_10.above():
            if value:
                self._pyobject.tp_flags |= TpFlags.IMMUTABLETYPE
            else:
                self._pyobject.tp_flags &= ~TpFlags.IMMUTABLETYPE
        else:  # pragma: no cover
            if value:
                self._pyobject.tp_flags &= ~TpFlags.HEAPTYPE
            else:
                self._pyobject.tp_flags |= TpFlags.HEAPTYPE

    @contextmanager
    def as_mutable(self) -> Generator[Self, None, None]:
        """Context manager to temporarily set the type as mutable."""
        if not self.immutable:
            yield self
            return
        self.immutable = False
        self._pyobject.Modified()
        yield self
        self.immutable = True
        self._pyobject.Modified()

    def _try_alloc(self, slot: Slot):
        # Check if there is a ptr class
        if slot.ptr_type is None:
            return
        # Check if the slot is a null pointer
        ptr = getattr(self._pyobject, slot.parts[0])
        if not ptr:
            # Make a new ptr type struct
            new = slot.ptr_type()
            ptr.contents = new

    def __getitem__(self, key: str) -> Any:
        """Get an attribute from the type object."""
        return self._pyobject.GetAttr(key)

    def __setitem__(self, key: str | tuple[str, ...], value: Any) -> None:
        """
        Set attributes on the type object.

        Multiple string keys can be used to set multiple attributes to the same value.
        """
        keys = (key,) if isinstance(key, str) else key
        for k in keys:
            # Cache original implementation
            base = self.base
            if not in_cache(base, k):
                if (attr := getattr(base, k, MISSING)) is not MISSING:
                    add_cache(base, k, attr)
            # Check if this is a slots attr
            if slot := get_slot(k):
                # Allocate sub-struct if needed
                self._try_alloc(slot)

            with self.as_mutable():
                self._pyobject.setattr_safe(k, value)

    # <-- Begin Managed::Properties (structs::py_type.PyTypeObject) -->

    # <-- End Managed::Properties -->

    def __getattr__(self, item: str):
        # Forward `tp_` attributes from PyTypeObject
        if item.startswith("tp_"):
            return getattr(self._pyobject, item)
        raise AttributeError(item)

    def __setattr__(self, key: str, value: Any) -> None:
        # Forward `tp_` attributes from PyTypeObject
        if key.startswith("tp_") and hasattr(self._pyobject, key):
            if not self._unsafe:
                raise UnsafeError(
                    f"Setting attribute {key} requires an unsafe context."
                )
            setattr(self._pyobject, key, value)
            return
        super().__setattr__(key, value)



from __future__ import annotations

import ctypes
from collections.abc import Sequence
from ctypes import POINTER, pointer, pythonapi, sizeof
from typing import Any, TypeVar, overload

from einspect.api import Py_ssize_t, seq_to_array
from einspect.protocols.delayed_bind import bind_api
from einspect.structs.deco import struct
from einspect.structs.py_object import Fields, PyObject, PyVarObject
from einspect.structs.traits import IsGC
from einspect.types import Array, ptr

_VT = TypeVar("_VT")


@struct
class PyTupleObject(PyVarObject[tuple, None, _VT], IsGC):
    """
    Defines a PyTupleObject Structure.

    https://github.com/python/cpython/blob/3.11/Include/cpython/tupleobject.h
    """

    # Size of this array is only known after creation
    # this field just serves to get the address of the array
    _ob_item_0: ptr[PyObject] * 0

    def _format_fields_(self) -> Fields:
        return {**super()._format_fields_(), "ob_item": "Array[*PyObject]"}

    @overload
    @classmethod
    def from_object(cls, obj: tuple[_VT, ...]) -> PyTupleObject[_VT]:
        ...

    @overload
    @classmethod
    def from_object(cls, obj: tuple[...]) -> PyTupleObject[Any]:
        ...

    @classmethod
    def from_object(cls, obj: tuple[_VT, ...]) -> PyTupleObject[_VT]:
        """Create a PyTupleObject from an object."""
        return super().from_object(obj)  # type: ignore

    @property
    def mem_size(self) -> int:
        """Return the size of the PyObject in memory."""
        # Need to add size * ob_size to our base size
        base = super().mem_size
        int_size = ctypes.sizeof(Py_ssize_t)
        return base + (int_size * self.ob_size)

    @property
    def ob_item(self) -> Array[ptr[PyObject]]:
        """Return the ob_item field."""
        items_addr = ctypes.addressof(self._ob_item_0)
        arr = POINTER(PyObject) * self.ob_size
        return arr.from_address(items_addr)

    @ob_item.setter
    def ob_item(self, value: Array[ptr[PyObject]] | Sequence[ptr[PyObject]]) -> None:
        """Set the ob_item field."""
        arr = seq_to_array(value, ptr[PyObject])
        items_addr = ctypes.addressof(self._ob_item_0)
        ctypes.memmove(items_addr, arr, sizeof(arr))

    @bind_api(pythonapi["PyTuple_GetItem"])
    def GetItem(self, index: int) -> pointer[PyObject[_VT, None, None]]:
        """Return the item at the given index."""

    @bind_api(pythonapi["PyTuple_GetSlice"])
    def GetSlice(self, start: int, stop: int) -> pointer[PyTupleObject[_VT]]:
        """Return a slice of the tuple."""

    @bind_api(pythonapi["PyTuple_SetItem"])
    def SetItem(self, index: int, value: object) -> int:
        """
        Set a value to a given index.

        - Can only be used when refcount is equal to 1.
        """

    @bind_api(pythonapi["_PyTuple_Resize"])
    def Resize(self, size: int) -> None:
        """Resize the tuple to the given size."""



from einspect.views.view_base import AnyView, VarView, View
from einspect.views.view_bool import BoolView
from einspect.views.view_cfunction import CFunctionView
from einspect.views.view_dict import DictView
from einspect.views.view_float import FloatView
from einspect.views.view_function import FunctionView
from einspect.views.view_int import IntView
from einspect.views.view_list import ListView
from einspect.views.view_mapping_proxy import MappingProxyView
from einspect.views.view_set import SetView
from einspect.views.view_str import StrView
from einspect.views.view_tuple import TupleView
from einspect.views.view_type import TypeView

__all__ = (
    "AnyView",
    "VarView",
    "View",
    "BoolView",
    "DictView",
    "FloatView",
    "FunctionView",
    "CFunctionView",
    "IntView",
    "ListView",
    "MappingProxyView",
    "StrView",
    "SetView",
    "TupleView",
    "TypeView",
)



"""https://github.com/python/cpython/blob/3.11/Include/object.h"""
from __future__ import annotations

from ctypes import PYFUNCTYPE, Structure, c_char_p, c_int, c_void_p, py_object
from enum import IntEnum

from einspect.api import Py_ssize_t
from einspect.structs.deco import struct
from einspect.structs.include.pybuffer_h import getbufferproc, releasebufferproc
from einspect.types import ptr

__all__ = (
    "unaryfunc",
    "binaryfunc",
    "ternaryfunc",
    "inquiry",
    "lenfunc",
    "ssizeargfunc",
    "ssizessizeargfunc",
    "ssizeobjargproc",
    "ssizessizeobjargproc",
    "objobjargproc",
    "objobjproc",
    "visitproc",
    "traverseproc",
    "freefunc",
    "destructor",
    "getattrfunc",
    "getattrofunc",
    "setattrfunc",
    "setattrofunc",
    "reprfunc",
    "hashfunc",
    "richcmpfunc",
    "getiterfunc",
    "iternextfunc",
    "descrgetfunc",
    "descrsetfunc",
    "initproc",
    "newfunc",
    "allocfunc",
    "vectorcallfunc",
    "sendfunc",
    "PySendResult",
    "TpFlags",
    "PyBufferProcs",
    "PyAsyncMethods",
    "PyNumberMethods",
    "PyMappingMethods",
    "PySequenceMethods",
)

# Function types
# https://github.com/python/cpython/blob/3.11/Include/object.h#L196-L227
unaryfunc = PYFUNCTYPE(py_object, py_object)
binaryfunc = PYFUNCTYPE(py_object, py_object, py_object)
ternaryfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
inquiry = PYFUNCTYPE(c_int, py_object)
lenfunc = PYFUNCTYPE(Py_ssize_t, py_object)
ssizeargfunc = PYFUNCTYPE(py_object, py_object, Py_ssize_t)
ssizessizeargfunc = PYFUNCTYPE(py_object, py_object, Py_ssize_t, Py_ssize_t)
ssizeobjargproc = PYFUNCTYPE(c_int, py_object, Py_ssize_t, py_object)
ssizessizeobjargproc = PYFUNCTYPE(c_int, py_object, Py_ssize_t, Py_ssize_t, py_object)
objobjargproc = PYFUNCTYPE(c_int, py_object, py_object, py_object)

objobjproc = PYFUNCTYPE(c_int, py_object, py_object)
visitproc = PYFUNCTYPE(c_int, py_object, c_void_p)
traverseproc = PYFUNCTYPE(c_int, py_object, visitproc, c_void_p)

freefunc = PYFUNCTYPE(None, c_void_p)
destructor = PYFUNCTYPE(None, py_object)
getattrfunc = PYFUNCTYPE(py_object, py_object, c_char_p)
getattrofunc = PYFUNCTYPE(py_object, py_object, py_object)
setattrfunc = PYFUNCTYPE(c_int, py_object, c_char_p, py_object)
setattrofunc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
reprfunc = PYFUNCTYPE(py_object, py_object)
hashfunc = PYFUNCTYPE(Py_ssize_t, py_object)
richcmpfunc = PYFUNCTYPE(py_object, py_object, py_object, c_int)
getiterfunc = PYFUNCTYPE(py_object, py_object)
iternextfunc = PYFUNCTYPE(py_object, py_object)

descrgetfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
descrsetfunc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
initproc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
newfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
allocfunc = PYFUNCTYPE(py_object, py_object, Py_ssize_t)

# PyObject *(*vectorcallfunc)(PyObject *callable, PyObject *const *args, size_t nargsf, PyObject *kwnames)
vectorcallfunc = PYFUNCTYPE(py_object, py_object, ptr[py_object], Py_ssize_t, py_object)
# PySendResult (*sendfunc)(PyObject *iter, PyObject *value, PyObject **result)
sendfunc = PYFUNCTYPE(c_int, py_object, py_object, ptr[py_object])


class PySendResult(IntEnum):
    """Result of calling PyIter_Send."""

    PYGEN_RETURN = 0
    PYGEN_ERROR = -1
    PYGEN_NEXT = 1


class TpFlags(IntEnum):
    """
    Type flags (tp_flags)

    https://github.com/python/cpython/blob/3.11/Include/object.h#L350-L440
    """

    STATIC_BUILTIN = 1 << 1
    MANAGED_WEAKREF = 1 << 3
    MANAGED_DICT = 1 << 4
    PREHEADER = MANAGED_WEAKREF | MANAGED_DICT
    # Set if instances of the type object are treated as sequences for pattern matching
    SEQUENCE = 1 << 5
    # Set if instances of the type object are treated as mappings for pattern matching
    MAPPING = 1 << 6
    # Disallow creating instances of the type: set tp_new to NULL and don't create
    # the "__new__" key in the type dictionary.
    DISALLOW_INSTANTIATION = 1 << 7
    # Set if the type object is immutable: type attributes cannot be set nor deleted
    IMMUTABLETYPE = 1 << 8
    # Set if the type object is dynamically allocated
    HEAPTYPE = 1 << 9
    # Set if the type allows subclassing
    BASETYPE = 1 << 10
    # Set if the type implements the vectorcall protocol (PEP 590)
    HAVE_VECTORCALL = 1 << 11
    # Set if the type is 'ready' -- fully initialized
    READY = 1 << 12
    # Set while the type is being 'readied', to prevent recursive ready calls
    READYING = 1 << 13
    # Objects support garbage collection (see objimpl.h)
    HAVE_GC = 1 << 14
    # These two bits are preserved for Stackless Python, next after this is 17
    HAVE_STACKLESS_EXTENSION = 0  # if STACKLESS (3 << 15)
    # Objects behave like an unbound method
    METHOD_DESCRIPTOR = 1 << 17
    # Object has up-to-date type attribute cache
    VALID_VERSION_TAG = 1 << 19
    # Type is abstract and cannot be instantiated
    IS_ABSTRACT = 1 << 20
    # This undocumented flag gives certain built-ins their unique pattern-matching
    # behavior, which allows a single positional subpattern to match against the
    # subject itself (rather than a mapped attribute on it):
    MATCH_SELF = 1 << 22
    # These flags are used to determine if a type is a subclass.
    LONG_SUBCLASS = 1 << 24
    LIST_SUBCLASS = 1 << 25
    TUPLE_SUBCLASS = 1 << 26
    BYTES_SUBCLASS = 1 << 27
    UNICODE_SUBCLASS = 1 << 28
    DICT_SUBCLASS = 1 << 29
    BASE_EXC_SUBCLASS = 1 << 30
    TYPE_SUBCLASS = 1 << 31
    DEFAULT = HAVE_STACKLESS_EXTENSION | 0
    # Some of these flags reuse lower bits (removed as part of the Python 3.0 transition).
    HAVE_FINALIZE = 1 << 0
    HAVE_VERSION_TAG = 1 << 18


@struct
class PyAsyncMethods(Structure):
    am_await: unaryfunc
    am_aiter: unaryfunc
    am_anext: unaryfunc
    am_send: sendfunc


@struct
class PyBufferProcs(Structure):
    bf_getbuffer: getbufferproc
    bf_releasebuffer: releasebufferproc


@struct
class PyNumberMethods(Structure):
    """
    Number implementations must check *both*
    arguments for proper type and implement the necessary conversions
    in the slot functions themselves.
    """

    nb_add: binaryfunc
    nb_subtract: binaryfunc
    nb_multiply: binaryfunc
    nb_remainder: binaryfunc
    nb_divmod: binaryfunc
    nb_power: ternaryfunc
    nb_negative: unaryfunc
    nb_positive: unaryfunc
    nb_absolute: unaryfunc
    nb_bool: inquiry
    nb_invert: unaryfunc
    nb_lshift: binaryfunc
    nb_rshift: binaryfunc
    nb_and: binaryfunc
    nb_xor: binaryfunc
    nb_or: binaryfunc
    nb_int: unaryfunc
    nb_reserved: c_void_p  # formerly nb_long
    nb_float: unaryfunc

    nb_inplace_add: binaryfunc
    nb_inplace_subtract: binaryfunc
    nb_inplace_multiply: binaryfunc
    nb_inplace_remainder: binaryfunc
    nb_inplace_power: ternaryfunc
    nb_inplace_lshift: binaryfunc
    nb_inplace_rshift: binaryfunc
    nb_inplace_and: binaryfunc
    nb_inplace_xor: binaryfunc
    nb_inplace_or: binaryfunc

    nb_floor_divide: binaryfunc
    nb_true_divide: binaryfunc
    nb_inplace_floor_divide: binaryfunc
    nb_inplace_true_divide: binaryfunc

    nb_index: unaryfunc

    nb_matrix_multiply: binaryfunc
    nb_inplace_matrix_multiply: binaryfunc


@struct
class PyMappingMethods(Structure):
    mp_length: lenfunc
    mp_subscript: binaryfunc
    mp_ass_subscript: objobjargproc


@struct
class PySequenceMethods(Structure):
    sq_length: lenfunc
    sq_concat: binaryfunc
    sq_repeat: ssizeargfunc
    sq_item: ssizeargfunc
    was_sq_slice: c_void_p
    sq_ass_item: ssizeobjargproc
    was_sq_ass_slice: c_void_p
    sq_contains: objobjproc
    sq_inplace_concat: binaryfunc
    sq_inplace_repeat: ssizeargfunc
